/*
 *  Driver for AMD InterWave soundcard
 *  Copyright (c) by Jaroslav Kysela (Perex soft)
 */

#define ULTRA_MAIN_OBJECT_FILE
#include "driver.h"
#include "pcm.h"
#include "mixer.h"
#include "gus.h"
#include "cs4231.h"

char *ultra_id = NULL;		/* ID for this card */
int ultra_port = ULTRA_AUTO;	/* 0x210,0x220,0x230,0x240,0x250,0x260 */
int ultra_irq = ULTRA_AUTO;	/* 0,1,3,5,9,11,12,15 */
int ultra_dma1 = ULTRA_AUTO;	/* 0,1,3,5,6,7 */
int ultra_dma2 = ULTRA_AUTO;	/* 0,1,3,5,6,7 (-1 = disable) */
int ultra_dma1_size = ULTRA_AUTO; /* 8,16,32,64,128 */
int ultra_dma2_size = ULTRA_AUTO; /* 8,16,32,64,128 */
int ultra_joystick_dac = 29;	/* 0 to 31, (0.59V-4.52V or 0.389V-2.98V) */
#ifdef MODULE_PARM
MODULE_PARM( ultra_id, "s" );
MODULE_PARM_DESC( ultra_port, "ID string for InterWave soundcard." );
MODULE_PARM( ultra_port, "i" );
MODULE_PARM_DESC( ultra_port, "Port # for InterWave driver." );
MODULE_PARM( ultra_irq, "i" );
MODULE_PARM_DESC( ultra_irq, "IRQ # for InterWave driver." );
MODULE_PARM( ultra_dma1, "i" );
MODULE_PARM_DESC( ultra_dma1, "DMA1 # for InterWave driver." );
MODULE_PARM( ultra_dma2, "i" );
MODULE_PARM_DESC( ultra_dma2, "DMA2 # for InterWave driver." );
MODULE_PARM( ultra_dma1_size, "i" );
MODULE_PARM_DESC( ultra_dma1_size, "DMA1 size in kB for InterWave driver." );
MODULE_PARM( ultra_dma2_size, "i" );
MODULE_PARM_DESC( ultra_dma2_size, "DMA2 size in kB for InterWave driver." );
MODULE_PARM( ultra_joystick_dac, "i" );
MODULE_PARM_DESC( ultra_joystick_dac, "Joystick DAC level 0.59V-4.52V or 0.389V-2.98V for InterWave driver." );
#endif

static ultra_card_t *ultra_card = NULL;
static ultra_gus_card_t *ultra_gus = NULL;
static ultra_pcm_t *ultra_pcm = NULL;
static ultra_mixer_t *ultra_mixer = NULL;

static unsigned short ultra_gus_status;
static unsigned short ultra_pcm_status;

static void ultra_interwave_use_inc( void )
{
  MOD_INC_USE_COUNT;
}

static void ultra_interwave_use_dec( void )
{
  MOD_DEC_USE_COUNT;
}

static int ultra_interwave_detect( ultra_gus_card_t *gus, unsigned short port )
{
  unsigned long flags;
  unsigned char rev1, rev2;

  if ( ultra_register_ioport( gus -> card, port, 16, "InterWave (Adlib/SB compatibility)" ) < 0 )
    return -EBUSY;
  if ( ultra_register_ioport( gus -> card, port + 0x100, 12, "InterWave (synthesizer)" ) < 0 ) {
    ultra_unregister_ioports( gus -> card );
    return -EBUSY;
  }
  if ( ultra_register_ioport( gus -> card, port + 0x10c, 4, "InterWave (CS4231)" ) < 0 ) {
    ultra_unregister_ioports( gus -> card );
    return -EBUSY;
  }
  ultra_gus_set_port( gus, port );
  ultra_gf1_i_write8( gus, ULTRA_GF1_GB_RESET, 0 );	/* reset GF1 */
#ifdef ULTRACFG_DEBUG_DETECT
  if ( ((d = ultra_gf1_i_look8( gus, ULTRA_GF1_GB_RESET )) & 0x07) != 0 ) {
    PRINTK( "ultra: [0x%x] check 1 failed - 0x%x\n", gus -> gf1.port, d );
    goto __nodev;
  }
#else
  if ( (ultra_gf1_i_look8( gus, ULTRA_GF1_GB_RESET ) & 0x07) != 0 ) goto __nodev;
#endif
  ultra_delay( 16 );
  ultra_gf1_i_write8( gus, ULTRA_GF1_GB_RESET, 1 );	/* release reset */
  ultra_delay( 16 );
#ifdef ULTRACFG_DEBUG_DETECT
  if ( ((d = ultra_gf1_i_look8( gus, ULTRA_GF1_GB_RESET )) & 0x07) != 1 ) {
    PRINTK( "ultra: [0x%x] check 2 failed - 0x%x\n", gus -> gf1.port, d );
    goto __nodev;
  }
#else
  if ( (ultra_gf1_i_look8( gus, ULTRA_GF1_GB_RESET ) & 0x07) != 1 ) goto __nodev;
#endif

  CLI( &flags );
  rev1 = ultra_gf1_look8( gus, ULTRA_GF1_GB_VERSION_NUMBER );
  ultra_gf1_write8( gus, ULTRA_GF1_GB_VERSION_NUMBER, ~rev1 );
  rev2 = ultra_gf1_look8( gus, ULTRA_GF1_GB_VERSION_NUMBER );
  ultra_gf1_write8( gus, ULTRA_GF1_GB_VERSION_NUMBER, rev1 );
  STI( &flags );
#ifdef ULTRACFG_DEBUG_DETECT
  PRINTK( "ultra: [0x%x] InterWave check - rev1=0x%x, rev2=0x%x\n", gus -> gf1.port, rev1, rev2 );
#endif
  if ( ( rev1 & 0xf0 ) == ( rev2 & 0xf0 ) &&
       ( rev1 & 0x0f ) != ( rev2 & 0x0f ) )
    {
#ifdef ULTRACFG_DEBUG_DETECT
      PRINTK( "ultra: [0x%x] InterWave check - passed\n", gus -> gf1.port );
#endif
      gus -> interwave = 1;
      return 0;			/* ok.. We have InterWave board */
    }
#ifdef ULTRACFG_DEBUG_DETECT
  PRINTK( "ultra: [0x%x] InterWave check - failed\n", gus -> gf1.port );
#endif
  __nodev:
  ultra_unregister_ioports( gus -> card );
  return -ENODEV;
}

static void ultra_interwave_interrupt( int irq, void *dev_id, struct pt_regs *regs )
{
  register int loop;
  register unsigned short status;

  do {
    loop = 0;
    if ( (status = INB( ultra_gus_status )) ) {
      ultra_gus_interrupt( ultra_gus, status );
      loop++;
    }
    status = INB( ultra_pcm_status );
    if ( (status = INB( ultra_pcm_status )) & 0x01 ) {	/* IRQ bit is set? */
      ultra_cs4231_interrupt( ultra_pcm, status );
      loop++;
    }
  } while ( loop );
}

static int ultra_interwave_resources( ultra_gus_card_t *gus )
{
  static int possible_irqs[] = { 5, 11, 12, 9, 7, 15, 3, 4, -1 };
  static int possible_dmas[] = { 0, 1, 3, 5, 6, 7, -1 };
  ultra_card_t *card;
  int err;

  card = gus -> card;

  if ( (err = ultra_register_interrupt( card, ultra_irq, ultra_interwave_interrupt, "InterWave", possible_irqs )) < 0 )
    return err;
  if ( (err = ultra_register_dma_channel( card, ULTRA_GUS_DMA_CRECORD, "GFA1/CS4231 record", ultra_dma1, ultra_dma1_size, possible_dmas )) < 0 ) {
    return err;
  }
  if ( ultra_dma2 >= 0 ) {
    if ( (err = ultra_register_dma_channel( card, ULTRA_GUS_DMA_CPLAY, "CS4231 playback", ultra_dma2, ultra_dma2_size, possible_dmas )) < 0 ) {
      return err;
    }
  }
  return 0;
}

static void ultra_interwave_reset( ultra_gus_card_t *gus )
{
  ultra_gf1_write8( gus, ULTRA_GF1_GB_RESET, 0x00 );
  ultra_delay( 16 );
  ultra_gf1_write8( gus, ULTRA_GF1_GB_RESET, 0x01 );
  ultra_delay( 16 );
}

static void ultra_interwave_bank_sizes( ultra_gus_card_t *gus, int *sizes )
{
  int idx;
  unsigned int local;
  unsigned char d;

  for ( idx = 0; idx < 4; idx++ )
    {
      sizes[ idx ] = 0;
      d = 0x55;
      for ( local = idx << 22;
            local < ( idx << 22 ) + 0x400000;
            local += 0x40000, d++ )
        {
          ultra_gf1_poke( gus, local, d );
          ultra_gf1_poke( gus, local + 1, d + 1 );
#if 0
          printk( "d = 0x%x, local = 0x%x, local + 1 = 0x%x, idx << 22 = 0x%x\n",
          		d,
                        ultra_gf1_peek( gus, local ),
                        ultra_gf1_peek( gus, local + 1 ),
                        ultra_gf1_peek( gus, idx << 22 ) );
#endif
          if ( ultra_gf1_peek( gus, local ) != d ||
               ultra_gf1_peek( gus, local + 1 ) != d + 1 ||
               ultra_gf1_peek( gus, idx << 22 ) != 0x55 ) break;
          sizes[ idx ]++;
        }
    }
#if 0
  printk( "sizes: %i %i %i %i\n", sizes[ 0 ], sizes[ 1 ], sizes[ 2 ], sizes[ 3 ] );
#endif
}

struct rom_hdr {
  /* 000 */ unsigned char iwave[ 8 ];
  /* 008 */ unsigned char rom_hdr_revision;
  /* 009 */ unsigned char series_number;
  /* 010 */ unsigned char series_name[ 16 ];
  /* 026 */ unsigned char date[ 10 ];
  /* 036 */ unsigned short vendor_revision_major;
  /* 038 */ unsigned short vendor_revision_minor;
  /* 040 */ unsigned int rom_size;
  /* 044 */ unsigned char copyright[ 128 ];
  /* 172 */ unsigned char vendor_name[ 64 ];
  /* 236 */ unsigned char rom_description[ 128 ];
  /* 364 */ unsigned char pad[ 147 ];
  /* 511 */ unsigned char csum;
};

static void ultra_interwave_detect_memory( ultra_gus_card_t *gus )
{
  static unsigned int lmc[ 13 ] = {
    0x00000001, 0x00000101, 0x01010101, 0x00000401,
    0x04040401, 0x00040101, 0x04040101, 0x00000004,
    0x00000404, 0x04040404, 0x00000010, 0x00001010,
    0x10101010
  };

  int i, bank_pos, pages;
  unsigned int lmct;
  int psizes[ 4 ];
  unsigned char csum;
  struct rom_hdr romh;

  gus -> card -> version = ULTRA_CARD_TYPE_GUS_PNP;
  ultra_interwave_reset( gus );
  ultra_gf1_write8( gus, ULTRA_GF1_GB_GLOBAL_MODE, ultra_gf1_read8( gus, ULTRA_GF1_GB_GLOBAL_MODE ) | 0x01 );	/* enhanced mode */
  ultra_gf1_write8( gus, ULTRA_GF1_GB_MEMORY_CONTROL, 0x01 );  /* DRAM I/O cycles selected */
  ultra_gf1_write16( gus, ULTRA_GF1_GW_MEMORY_CONFIG, ( ultra_gf1_look16( gus, ULTRA_GF1_GW_MEMORY_CONFIG ) & 0xff10 ) | 0x004c );
  /* ok.. simple test of memory size */
  pages = 0;
  ultra_gf1_poke( gus, 0, 0x55 );
  ultra_gf1_poke( gus, 1, 0xaa );
#if 1
  if ( ultra_gf1_peek( gus, 0 ) == 0x55 && ultra_gf1_peek( gus, 1 ) == 0xaa )
#else
  if ( 0 )		/* ok.. for testing of 0k RAM */
#endif
    {
      ultra_interwave_bank_sizes( gus, psizes );
      lmct = ( psizes[ 3 ] << 24 ) | ( psizes[ 2 ] << 16 ) |
             ( psizes[ 1 ] << 8 ) | psizes[ 0 ];
#if 0
      printk( "lmct = 0x%08x\n", lmct );
#endif
      for ( i = 0; i < sizeof( lmc ) / sizeof( unsigned int ); i++ )
        if ( lmct == lmc[ i ] )
          {
#if 0
            printk( "found !!! %i\n", i );
#endif
            ultra_gf1_write16( gus, ULTRA_GF1_GW_MEMORY_CONFIG, ( ultra_gf1_look16( gus, ULTRA_GF1_GW_MEMORY_CONFIG ) & 0xfff0 ) | i );
            ultra_interwave_bank_sizes( gus, psizes );
            break;
          }
      if ( i >= sizeof( lmc ) / sizeof( unsigned int ) && !gus -> gf1.enh_mode )
        ultra_gf1_write16( gus, ULTRA_GF1_GW_MEMORY_CONFIG, ( ultra_gf1_look16( gus, ULTRA_GF1_GW_MEMORY_CONFIG ) & 0xfff0 ) | 2 );
      for ( i = 0; i < 4; i++ )
        {
          gus -> gf1.banks_8[ i ].address =
          gus -> gf1.banks_16[ i ].address = i << 22;
          gus -> gf1.banks_8[ i ].size =
          gus -> gf1.banks_16[ i ].size = psizes[ i ] << 18;
          pages += psizes[ i ];
        }
    }
  pages <<= 18;
  gus -> gf1.memory = pages;
  
  ultra_gf1_write8( gus, ULTRA_GF1_GB_MEMORY_CONTROL, 0x03 );  /* select ROM */
  ultra_gf1_write16( gus, ULTRA_GF1_GW_MEMORY_CONFIG, ( ultra_gf1_look16( gus, ULTRA_GF1_GW_MEMORY_CONFIG ) & 0xff1f ) | ( 4 << 5 ) );
  gus -> gf1.rom_banks = 0;
  gus -> gf1.rom_memory = 0;
  for ( bank_pos = 0; bank_pos < 16L * 1024L * 1024L; bank_pos += 4L * 1024L * 1024L )
    {
      for ( i = 0; i < sizeof( struct rom_hdr ); i++ )
        *(((unsigned char *)&romh) + i) = ultra_gf1_peek( gus, i + bank_pos );
#ifdef ULTRACFG_DEBUG_ROM
      printk( "ultra: ROM at 0x%06x = %02x:%02x:%02x:%02x:%02x:%02x:%02x:%02x\n", bank_pos,
        		romh.iwave[ 0 ], romh.iwave[ 1 ], romh.iwave[ 2 ], romh.iwave[ 3 ],
        		romh.iwave[ 4 ], romh.iwave[ 5 ], romh.iwave[ 6 ], romh.iwave[ 7 ] );
#endif
      if ( strncmp( romh.iwave, "INTRWAVE", 8 ) ) continue;  /* first check */
      csum = 0;
      for ( i = 0; i < sizeof( struct rom_hdr ) - 1; i++ )
        csum += *(((unsigned char *)&romh) + i); 
#ifdef ULTRACFG_DEBUG_ROM
      printk( "ultra: ROM checksum = 0x%x == 0x%x (computed)\n", romh.csum, (unsigned char)(256 - csum) );
#endif
      if ( 256 - csum != romh.csum ) continue;		  /* not valid rom */
      gus -> gf1.rom_banks++;
      gus -> gf1.rom_present |= 1 << ( bank_pos >> 22 );
      gus -> gf1.rom_memory = ultra_get_dword( (char *)&romh.rom_size, 0 );
    }
  if ( gus -> gf1.rom_memory > 0 )
    {
      if ( gus -> gf1.rom_banks == 1 && gus -> gf1.rom_present == 8 )
        gus -> card -> version = ULTRA_CARD_TYPE_IW_DYNASONIC;
    }
  ultra_gf1_write8( gus, ULTRA_GF1_GB_MEMORY_CONTROL, 0x00 );  /* select RAM */
    
  if ( !gus -> gf1.enh_mode ) ultra_interwave_reset( gus );
}

static void ultra_interwave_init( ultra_gus_card_t *gus )
{
  unsigned long flags;

  /* ok.. some InterWave specific initialization */
  CLI( &flags );
  ultra_gf1_write8( gus, ULTRA_GF1_GB_SOUND_BLASTER_CONTROL, 0x00 );
  ultra_gf1_write8( gus, ULTRA_GF1_GB_COMPATIBILITY, 0x1f );
  ultra_gf1_write8( gus, ULTRA_GF1_GB_DECODE_CONTROL, 0x49 );
  ultra_gf1_write8( gus, ULTRA_GF1_GB_VERSION_NUMBER, 0x11 );
  ultra_gf1_write8( gus, ULTRA_GF1_GB_MPU401_CONTROL_A, 0x00 );
  ultra_gf1_write8( gus, ULTRA_GF1_GB_MPU401_CONTROL_B, 0x30 );
  ultra_gf1_write8( gus, ULTRA_GF1_GB_EMULATION_IRQ, 0x00 );
  STI( &flags );
  gus -> equal_irq = 1;
  if ( ultra_dma2 < 0 ) gus -> equal_dma = 1;
  gus -> codec_flag = 1;
  gus -> interwave = 1;
  gus -> max_flag = 1;
  gus -> joystick_dac = ultra_joystick_dac;
  gus -> gf1.irq = gus -> card -> irqs[ ULTRA_GUS_IRQ_GF1 ].number;
}

#define CS4231_PRIVATE( left, right, shift, mute ) ((left << 24)|(right << 16)|(shift<<8)|mute)

static int ultra_interwave_mixer( ultra_mixer_t *mixer )
{
  static struct ULTRA_STRU_MIXER_CHANNEL_HW master = {
    ULTRA_MIXER_PRI_MASTER,		/* priority */
    ULTRA_MIXER_ID_MASTER,		/* device name */
    ULTRA_MIXER_VOLUME,			/* OSS device # */
    1, 1, 1,				/* mute/stereo/record */
    31,					/* max. value */
    CS4231_PRIVATE( CS4231_LINE_LEFT_OUTPUT, CS4231_LINE_RIGHT_OUTPUT, 0, 0x80 ),
    NULL,				/* compute volume */
    NULL,				/* record source */
    NULL,				/* set mute */
    NULL,				/* set volume level */
  };
  ultra_mixer_channel_t *channel;
  int idx;
  
  /* ok. InterWave have MIC different (stereo) */
  channel = NULL;
  for ( idx = 0; idx < mixer -> channels_count; idx++ ) {
    channel = mixer -> channels[ idx ];
    switch ( channel -> hw.priority ) {
      case ULTRA_MIXER_PRI_MIC:
        channel -> hw.stereo = 1;
        channel -> hw.max = 31;
        channel -> hw.private_value = CS4231_PRIVATE( CS4231_LEFT_MIC_INPUT, CS4231_RIGHT_MIC_INPUT, 0, 0x80 );
        break;
      case ULTRA_MIXER_PRI_AUXA:
        channel -> hw.priority = ULTRA_MIXER_PRI_SYNTHESIZER;
        channel -> hw.ossdev = ULTRA_MIXER_SYNTH;
        strcpy( channel -> hw.name, ULTRA_MIXER_ID_SYNTHESIZER );
        ultra_mixer_reorder_channel( mixer, channel );
        break;
      case ULTRA_MIXER_PRI_AUXB:
        channel -> hw.priority = ULTRA_MIXER_PRI_CD;
        channel -> hw.ossdev = ULTRA_MIXER_CD;
        strcpy( channel -> hw.name, ULTRA_MIXER_ID_CD );
        ultra_mixer_reorder_channel( mixer, channel );
        break;
    }
  }
  if ( !channel ) return -EINVAL;
  /* add master volume control */
  master.set_record_source = channel -> hw.set_record_source;
  master.set_mute = channel -> hw.set_mute;
  master.set_volume_level = channel -> hw.set_volume_level;
  channel = ultra_mixer_new_channel( mixer, &master );
  if ( !channel ) return -ENOMEM;
  return 0;
}

int init_module( void )
{
  static int possible_ports[] = { 0x220,0x240,0x260,0x210,0x230,0x250, -1 };
  int *ports = possible_ports;
    
  ultra_card = ultra_card_new( ultra_id, "AMD InterWave", ultra_interwave_use_inc, ultra_interwave_use_dec );
  if ( !ultra_card ) return -ENOMEM;
  ultra_gus = ultra_gus_new_card( ultra_card );
  if ( !ultra_gus ) {
    ultra_card_free( ultra_card );
    return -ENOMEM;
  }
  if ( ultra_port == ULTRA_AUTO ) {
    for ( ports = possible_ports; *ports >= 0; ports++ ) {
      if ( !ultra_interwave_detect( ultra_gus, *ports ) ) break;
    }
    if ( *ports < 0 ) {
      ultra_card_free( ultra_card );
      PRINTK( "ultra: InterWave soundcard not found or device busy\n" );
      return -ENODEV;
    }
  } else {
    if ( ultra_interwave_detect( ultra_gus, ultra_port ) ) {
      ultra_card_free( ultra_card );
      PRINTK( "ultra: InterWave soundcard not found at 0x%x or device busy\n", ultra_port );
      return -ENODEV;
    }
  }
  if ( ultra_card_register( ultra_card ) ) {
    ultra_card_free( ultra_card );
    return -ENOMEM;
  }
  ultra_gus_status = ultra_gus -> gf1.reg_irqstat;
  ultra_pcm_status = ultra_gus -> gf1.port + 0x10c + 2;
  if ( ultra_interwave_resources( ultra_gus ) < 0 ) {
    ultra_card_free( ultra_card );
    return -EBUSY;
  }
  ultra_interwave_init( ultra_gus );
  ultra_interwave_detect_memory( ultra_gus );
  if ( ultra_gus_init_dma_irq( ultra_gus, 1 ) < 0 ) {
    ultra_card_free( ultra_card );
    return -EINVAL;
  }
  ultra_pcm = ultra_cs4231_new_device( ultra_card,
                                       ultra_gus -> gf1.port + 0x10c,
                                       ultra_gus -> gf1.irq,
                                       ultra_gus -> equal_dma ? ULTRA_GUS_DMA_CRECORD : ULTRA_GUS_DMA_CPLAY,
                                       ULTRA_GUS_DMA_CRECORD,
                                       CS4231_HW_INTERWAVE );
  if ( !ultra_pcm ) {
    __return1:
    ultra_gus_init_dma_irq( ultra_gus, 0 );
    ultra_card_free( ultra_card );
    return -ENXIO;
  }
  if ( ultra_cs4231_probe( ultra_pcm ) < 0 ) {
    PRINTD( "ultra: Oops, seems that InterWave card doesn't have CODEC CS4231?\n" );
    goto __return1;
  }
  if ( ultra_pcm_register( ultra_pcm, 0 ) < 0 ) goto __return1;
  ultra_mixer = ultra_cs4231_new_mixer( ultra_pcm );
  if ( !ultra_mixer ) {
    goto __return1;
  }
  if ( ultra_interwave_mixer( ultra_mixer ) < 0 ) {
    __return2:
    ultra_mixer_unregister( ultra_mixer );
    return -ENOMEM;
  }    
  if ( ultra_mixer_register( ultra_mixer, 0 ) < 0 ) {
    goto __return2;
  }
  ultra_mixer_channel_init( ultra_mixer, ULTRA_MIXER_PRI_MASTER, 70, 70, 0 );
  ultra_mixer_channel_init( ultra_mixer, ULTRA_MIXER_PRI_SYNTHESIZER, 75, 75, 0 );
  ultra_mixer_set_kernel_mute( ultra_mixer, ULTRA_MIXER_PRI_PCM, ULTRA_MIX_MUTE );
  ultra_mixer_channel_init( ultra_mixer, ULTRA_MIXER_PRI_PCM, 85, 85, 0 );
  ultra_mixer_channel_init( ultra_mixer, ULTRA_MIXER_PRI_LINE, 70, 70, ULTRA_MIXER_FLG_MUTE );
  ultra_mixer_channel_init( ultra_mixer, ULTRA_MIXER_PRI_MIC, 70, 70, ULTRA_MIXER_FLG_MUTE );
  ultra_mixer_channel_init( ultra_mixer, ULTRA_MIXER_PRI_CD, 70, 70, ULTRA_MIXER_FLG_MUTE );
  ultra_mixer_channel_init( ultra_mixer, ULTRA_MIXER_PRI_GAIN, 0, 0, 0 );
  ultra_mixer_channel_init( ultra_mixer, ULTRA_MIXER_PRI_LOOPBACK, 0, 0, ULTRA_MIXER_FLG_MUTE );
#if 0
  ultra_cs4231_debug( (cs4231_t *)ultra_pcm -> playback.hw.private_data );
#endif
  return 0;
}

void cleanup_module( void )
{
  if ( ultra_gus ) {
    ultra_gus_init_dma_irq( ultra_gus, 0 );
    if ( ultra_mixer )
      ultra_mixer_unregister( ultra_mixer );
    if ( ultra_pcm )
      ultra_pcm_unregister( ultra_pcm );
    if ( ultra_card )
      ultra_card_unregister( ultra_card );
  }
}
