/*
 *  Routines for driver control
 *  Copyright (c) by Jaroslav Kysela <perex@jcu.cz>
 */

#include "driver.h"

struct ULTRA_CONTROL {
  ultra_card_t *card;
};

typedef struct ULTRA_CONTROL ultra_control_t;

static int ultra_control_open( unsigned short minor, struct file *file )
{
  ultra_card_t *card;
  ultra_control_t *control;

  minor >>= 3;
  minor &= 0x0e;
  if ( !(card = ultra_cards[ minor ]) ) return -ENXIO;
  control = (ultra_control_t *)ultra_malloc( sizeof( ultra_control_t ) );
  if ( !control ) return -ENOMEM;
  control -> card = card;
  file -> private_data = control;
  MOD_INC_USE_COUNT;
  card -> use_inc();
  return 0;
}

static int ultra_control_release( unsigned short minor, struct file *file )
{
  ultra_control_t *control;

  if ( file -> private_data ) {
    control = (ultra_control_t *)file -> private_data;
    file -> private_data = NULL;
    control -> card -> use_dec();
  }
  MOD_DEC_USE_COUNT;
  return 0;
}

static int ultra_control_ioctl( struct file *file, unsigned int cmd, unsigned long arg )
{
  return -EIO;
}

/*
 *  INIT PART
 */

static int ultra_control_unregister_callback( void );

static struct ultra_stru_register_minor ultra_control_reg = {
  "control",

  1,					/* verify soundcard # */

  { ULTRA_MINOR_UCONTROL, -1 },
  { -1 },

  ultra_control_unregister_callback,	/* unregister */

  NULL,					/* lseek */
  NULL,					/* read */
  NULL,					/* write */
  ultra_control_open,			/* open */
  ultra_control_release,		/* release */
#ifdef ULTRA_POLL
  NULL,					/* poll */
#else
  NULL,					/* select */
#endif
  ultra_control_ioctl,			/* ioctl */
  NULL					/* mmap */
};

static int ultra_control_unregister_callback( void )
{
  int err;

  if ( ( err = ultra_unregister_minor( &ultra_control_reg ) ) < 0 )
    PRINTD( "ultra-control: unregister error %d\n", err );
  return err;
}
 
int ultra_control_register( void )
{
  int err;
  
  if ( ( err = ultra_register_minor( &ultra_control_reg ) ) < 0 )
    return err;
  return 0;
}
